# FormLogic

[![Travis CI](https://api.travis-ci.org/shlima/form_logic.png)](https://travis-ci.org/shlima/form_logic)

## Цели и задача

Задача данного гема — создание «сервисных» объектов форм для того, чтобы инкапсулировать логику
взаимодействия накольких объектов и избавиться от лишних зависимостей внутри моделей.

Базовый класс, `FormLogic::Base`, от которого вы наследуете свои сервисы, включает в себя поведения
`ActiveModel`, что делает его совместимым со всеми билдерами форм, дает возможность добавлять
валидации на виртуальные атрибуты, а так же сепаративно тестировать такой класс.

## API на основе примера

Пример: нам необходимо создать пользователя, при этом в форме добавить поле `bonus_code`, которое будет связано с
моделью `Bonus`, если пользователь ввел значение в это поле, необходимо найти соответствующий бонус,
и создать `BonusTransaction`, `BalanceTransaction`, и самого пользователя внутри его транзакции.

В противном случае, — показать ошибку валидации на поле `bonus_code`, если пользователь ввел в него данные,
но такого бонуса не существует или он инвалидный.

Пример кода контроллера:

```ruby
  def create
    @form = UserRegistrationForm.new(User.new(params[:user]))
    @form.bonus_code = params.fetch(:user, {})[:bonus_code]
    if @form.valid? && @form.submit!
      redirect_to root_url
    else
      render :new
    end
  end
```

Пример form-object:

```ruby
class UserRegistrationForm < FormLogic::Base

  ## `model_name` используется как неймспейс формы
  ## в параметрах, и в качестве ключа для I18n
  declare_model_name :user

  ## «Проперти» - это атрибуты моделей. При создании объекта,
  ## передайте в качестве параметров все необходимые модели, на основе
  ## имен кдассов которых будут созданы аксессоры, например:
  ##  
  ##   Form.new(User.new, Bonus.new)
  ##
  ## создаст два аксессора `user` и `bonus`.
  ##
  ## По-сути - это враппер над делегатором:
  property :email, :password, :password_confirmation, to: :user

  ## «Виртуальные» атрибуты (те, которые не привязаны к модели)
  ## По-сути - это враппер над `attr_accessor`:
  virtual :bonus_code

  ## При вызове метода `valid?` будут провалидированы все модели,
  ## переданные сервису при его инициализации, ошибки которых будут скопированы
  ## на сам объект сервиса.
  ##
  ## Если вам нужно добавить кастомную логику валидации определите метод
  ## `while_validation` который вызывается в процессе работы медоды `valid?`
  ##
  ## Так как сервисный объект включает в себя `ActiveModel` вы можете
  ## определить новые методы валидации для любых атрибутов так же, как делаете
  ## это для классов `ActiveRecord`, например:
  ##
  ##   validates :bonus_code, :presence => true
  def while_validation
    if bonus_code.present? && bonus.nil?
      errors.add(:bonus_code, :invalid)
    end
  end

  ## Этот метод приведен для примера, вам не обязательно его имплементировать.
  ## Вы можете определить методы `save` или `create`, чтобы сохранить совместимость
  ## с текущими формами и контроллерами.
  ##
  ## Также код этого метода, при желании, может быть вынесен к контроллер.
  def submit!
    user.transaction do
      BonusTransaction.create!
      user.create!
    end
  end

  def bonus
    @bonus ||= Bonus.registration.coded(bonus_code).first
  end
end
```

## Пример теста

```ruby
describe UserRegistrationForm do
  let!(:bonus) { create(:bonus, :registration) }
  let(:user)   { build(:user)                  }

  subject do
    described_class.new(user).tap do |form|
      form.bonus_code = bonus.code
    end
  end

  describe 'validation' do
    it { should be_valid }

    it 'validates bonus_code' do
      subject.bonus_code = 'FakeCode'
      expect(subject).to have(1).error_on(:bonus_code)
    end

    it 'validates customer' do
      subject.customer.email = 'fake-email'
      expect(subject).to have(1).error_on(:email)
    end
  end
end
```

При этом код самих DOM моделей остается чистым

## Installation

Add this line to your application's Gemfile:

    gem 'form_logic'

And then execute:

    $ bundle

Or install it yourself as:

    $ gem install form_logic
